require "lighting"
require "scene"
require "vectorfunctions"

shadowStencil = love.graphics.newStencil(function()
                                            local shadowGeometry = calculateShadows(lights.current, scene)
                                            for i=1,#shadowGeometry do
                                                love.graphics.polygon("fill", unpack(shadowGeometry[i]))
                                            end
                                         end)

--- calculates the shadow geometry
-- @param lightsource table with coordinates of lightsource in [1] and[2] (x and y)
-- @param geometry table of tables with convex polygons, vertices ordered counterclockwise
function calculateShadows(lightsource, geometry)
    local shadowGeometry = {} -- same architecture as geometry table, a shadow consists of normally four vertices with two coordinates
    local shadowLength = 10000--lightsource.range
    
    for i,curPolygon in pairs(geometry) do
        -- for each edge calculate if the edge is facing to the current lightsource or not
        if not curPolygon.castsNoShadow then 
            local edgeFacingTo = {}
            for j=1,#curPolygon,2 do
                -- [i] means x coordinate, [i+1] means y coordinate
                local indexOfNextVertex = (j+2)%#curPolygon
                local normal = {-curPolygon[indexOfNextVertex+1] + curPolygon[j+1], curPolygon[indexOfNextVertex] - curPolygon[j]}
                -- love.graphics.line(unpack(normal))
                local lightToPoint = {curPolygon[j] - lightsource.x, curPolygon[j+1] - lightsource.y}
                
                normal = normalize(normal)
                lightToPoint = normalize(lightToPoint)
                
                local dotProduct = dot(normal, lightToPoint)
                if dotProduct > 0 then table.insert(edgeFacingTo, "front")
                else table.insert(edgeFacingTo, "back") end
            end
            
            -- now calculate the shadow geometry
            local curShadowGeometry = {}
            for j,curFacing in pairs(edgeFacingTo) do
                local nextIndex = (j+1)%#edgeFacingTo; if nextIndex == 0 then nextIndex = #edgeFacingTo end
                -- if there are two neighboured edges that face in different directions then there must be a shadow
                if curFacing == "front" and edgeFacingTo[nextIndex] == "back" then 
                    curShadowGeometry[1] = curPolygon[nextIndex*2-1]
                    curShadowGeometry[2] = curPolygon[nextIndex*2]
                    
                    local lightVecFrontBack = normalize({curPolygon[nextIndex*2-1] - lightsource.x, curPolygon[nextIndex*2] - lightsource.y})
                    curShadowGeometry[3] = curShadowGeometry[1] + lightVecFrontBack[1]*shadowLength
                    curShadowGeometry[4] = curShadowGeometry[2] + lightVecFrontBack[2]*shadowLength
                    
                elseif curFacing == "back" and edgeFacingTo[nextIndex] == "front" then 
                    curShadowGeometry[7] = curPolygon[nextIndex*2-1]
                    curShadowGeometry[8] = curPolygon[nextIndex*2]
                    
                    local lightVecBackFront = normalize({curPolygon[nextIndex*2-1] - lightsource.x, curPolygon[nextIndex*2] - lightsource.y})
                    curShadowGeometry[5] = curShadowGeometry[7] + lightVecBackFront[1]*shadowLength
                    curShadowGeometry[6] = curShadowGeometry[8] + lightVecBackFront[2]*shadowLength
                end
            end
            if  curShadowGeometry[1]
            and curShadowGeometry[2]
            and curShadowGeometry[3]
            and curShadowGeometry[4]
            and curShadowGeometry[5]
            and curShadowGeometry[6]
            and curShadowGeometry[7]
            and curShadowGeometry[8]
            then
                shadowGeometry[#shadowGeometry+1] = curShadowGeometry
            end
        end
	end
    return shadowGeometry
end